import boto3
import subprocess
import json
import os
import uuid
import time
import csv

REGION_NAME='ap-northeast-1'
CONFIG_DIR="/etc/a610-thermometer/"
METRIC_NAMESPACE='A610-Thermometer'

class AwsInteract:
    def __init__(self):
        os.environ['AWS_DEFAULT_REGION'] = REGION_NAME
        self.get_credentials()
        self.iot = boto3.client('iot')
        self.iot_data = boto3.client('iot-data')
        self.iam = boto3.client('iam')
        self.cw = boto3.client('cloudwatch')
        self.lamb = boto3.client('lambda')
        self.gw = boto3.client('apigateway')
        self.sns = boto3.client('sns')

        '''
        armadillo_mac = subprocess.getoutput(
            """ip link show dev eth0 | grep link | awk '{print toupper($2)}' | tr -d ':'""")
        '''

        kotai = subprocess.getoutput("cat /sys/bus/nvmem/devices/imx-ocotp0/nvmem | \
                                    hexdump -v -s 0xa0 -n 8 -e '/4 \"%08X\"' | \
                                    cut -c 5-")
        if self.register_armadillo(kotai):
            self.attach_armadillo()

        self.create_cloudwatch_rule(kotai)
        self.init_cloudwatch()
        self.setup_update_shadow_lambda()

    def get_credentials(self):
        with open(CONFIG_DIR + "credentials.csv", "r") as scenario:
            header = next(csv.reader(scenario))
            scenario = csv.reader(scenario)
            for step in scenario:
                for col_type, value in zip(header, step):
                    if col_type == 'Access key ID':
                        self.accesskeyid = value
                        os.environ['AWS_ACCESS_KEY_ID'] = self.accesskeyid
                    elif col_type == 'Secret access key':
                        self.secretsccesskey = value
                        os.environ['AWS_SECRET_ACCESS_KEY'] = self.secretsccesskey
                    elif col_type == 'Console login link':
                        self.accountId = value.split('.')[0].split('/')[-1]

    def exists(self, name):
        list_of_things = self.iot.list_things()
        for thing in list_of_things['things']:
            if thing['thingName'] == name:
                return True
        while 'nextToken' in list_of_things:
            t = list_of_things['nextToken']
            list_of_things = self.iot.list_things(nextToken=t)
            for thing in list_of_things['things']:
                if thing['thingName'] == name:
                    return True
        return False

    def register_armadillo(self, kotai):
        self.iot.create_thing_type(thingTypeName='Armadillo-610')
        self.name = 'Armadillo-610_' + kotai
        if self.exists(self.name):
            print("This Armadillo-610 already exists")
            return False
        self.iot.create_thing(thingName=self.name, thingTypeName='Armadillo-610')
        result = self.iot.create_keys_and_certificate(setAsActive=True)
        self.cert_arn = result['certificateArn']

        cert_id = result['certificateId']
        cert_id_filename = CONFIG_DIR + 'a610_cert_id_file'
        with open(cert_id_filename, "w") as cert_id_file:
            print(cert_id, file=cert_id_file)
        cert_id_file_path = os.path.abspath(cert_id_filename)
        os.chmod(cert_id_file_path, 0o600)

        cert_pem = result['certificatePem']
        cert_pem_filename = CONFIG_DIR + 'certificate.pem.crt'
        with open(cert_pem_filename, "w") as cert_pem_file:
            print(cert_pem, file=cert_pem_file)
        cert_pem_file_path = os.path.abspath(cert_pem_filename)
        os.chmod(cert_pem_file_path, 0o600)

        private_key_pem = result['keyPair']['PrivateKey']
        private_key_pem_filename = CONFIG_DIR + 'private.pem.key'
        with open(private_key_pem_filename, "w") as private_key_pem_file:
            print(private_key_pem, file=private_key_pem_file)
        private_key_pem_file_path = os.path.abspath(private_key_pem_filename)
        os.chmod(private_key_pem_file_path, 0o600)

        policy_document_filename = CONFIG_DIR + 'policy_document.templ'
        with open(policy_document_filename, "r") as policy_document_file:
            policy_document_text = policy_document_file.read()
        try:
            self.iot.create_policy(policyName='a610', policyDocument=policy_document_text)
        except:
            pass
        self.iot.attach_policy(policyName='a610', target=self.cert_arn)
        return True

    def attach_armadillo(self):
        self.iot.attach_thing_principal(thingName=self.name,
                                        principal=self.cert_arn)

    def update_shadow_reported(self, **kwargs):
        dic_payload = {}
        dic_payload.setdefault('state', {})
        dic_payload['state'].setdefault('reported', {})
        for key, val in kwargs.items():
            dic_payload['state']['reported'].setdefault(key, val)
        payload = json.dumps(dic_payload)
        self.iot_data.update_thing_shadow(thingName=self.name, payload=payload)

    def get_shadow_desired(self):
        shadow_stream = self.iot_data.get_thing_shadow(thingName=self.name)
        shadow_string = json.loads(shadow_stream['payload'].read().decode('utf-8'))
        shadow_string = shadow_string['state']['delta']
        return shadow_string

    def clear_shadow_desired(self):
        payload = json.dumps({'state': {'desired': None}})
        self.iot_data.update_thing_shadow(thingName=self.name, payload=payload)

    def exists_rule(self, name):
        list_of_rules = self.iot.list_topic_rules()
        for rule in list_of_rules['rules']:
            if rule['ruleName'] == name:
                return True
        while 'nextToken' in list_of_rules:
            t = list_of_rules['nextToken']
            list_of_rules = self.iot.list_topic_rules(nextToken=t)
            for rule in list_of_rules['rules']:
                if rule['ruleName'] == name:
                    return True
        return False

    def exists_role(self, name):
        list_of_roles = self.iam.list_roles()
        for role in list_of_roles['Roles']:
            if role['RoleName'] == name:
                return True
        while 'Marker' in list_of_roles:
            t = list_of_roles['Marker']
            list_of_roles = self.iam.list_roles(Marker=t)
            for role in list_of_roles['Roles']:
                if role['RoleName'] == name:
                    return True
        return False

    def create_role(self, name, roleTempl, policyTempl):
        with open(roleTempl, "r") as f:
            policy_doc = f.read()

        res_role = self.iam.create_role(
            RoleName=name,
            AssumeRolePolicyDocument=policy_doc
        )
        role_name = res_role['Role']['RoleName']

        with open(policyTempl, "r") as f:
            policy_doc = f.read()

        res_policy = self.iam.create_policy(
            PolicyName=name,
            PolicyDocument=policy_doc
        )
        policy_arn = res_policy['Policy']['Arn']

        self.iam.attach_role_policy(RoleName=role_name, PolicyArn=policy_arn)

        # ロールにポリシーをアタッチした後に反映まで時間がかかるが、
        # 2020/05/27現在、boto3でこれをwaitする手段はないためsleepする
        time.sleep(10)
        return res_role['Role']['Arn']

    def create_cloudwatch_rule(self, kotai):
        rule_name = kotai + '_cloudwatch_metrics'
        if self.exists_rule(rule_name):
            print("This rule already exists")
            return

        role_name = 'a610_thermometer_cloudwatch_metrics'
        if self.exists_role(role_name):
            res = self.iam.get_role(RoleName=role_name)
            role_arn = res['Role']['Arn']
        else:
            role_templ = CONFIG_DIR + "iot_rule.templ"
            policy_templ = CONFIG_DIR + "cw_putmetrics.templ"
            role_arn = self.create_role(
                name=role_name,
                roleTempl=role_templ,
                policyTempl=policy_templ
            )

        sql = 'SELECT state.reported.temp as temp FROM "$aws/things/' \
            + self.name + '/shadow/update/accepted"'
        self.iot.create_topic_rule(
            ruleName=rule_name,
            topicRulePayload={
                'sql': sql,
                'actions': [
                    {
                        'cloudwatchMetric': {
                            'roleArn': role_arn,
                            'metricNamespace': METRIC_NAMESPACE,
                            'metricName': self.name,
                            'metricValue': '${state.reported.temp}',
                            'metricUnit': 'None'
                        },
                    },
                ],
                'awsIotSqlVersion': '2016-03-23'
            }
        )

    def exists_dashboard(self, name):
        list_of_dashboards = self.cw.list_dashboards()
        for dashboard in list_of_dashboards['DashboardEntries']:
            if dashboard['DashboardName'] == name:
                return True
        while 'NextToken' in list_of_dashboards:
            t = list_of_dashboards['NextToken']
            list_of_dashboards = self.cw.list_dashboards(NextToken=t)
            for dashboard in list_of_dashboards['DashboardEntries']:
                if dashboard['DashboardName'] == name:
                    return True
        return False

    def delete_duplicate(self, arr):
        res = list(map(list, set(map(tuple, arr))))
        return res

    def add_param_to_widgets(self, body, namespace):
        append_list = [namespace, self.name]
        for index, widgets in enumerate(body['widgets']):
            if widgets['properties']['view'] == 'timeSeries' \
               or widgets['properties']['view'] == 'singleValue':
                arr = body['widgets'][index]['properties']['metrics']
                arr.append(append_list)
                res = self.delete_duplicate(arr)
                body['widgets'][index]['properties']['metrics'] = res
        return body

    def get_cloudwatch_dashboard(self, dashboardname):
        if self.exists_dashboard(dashboardname):
            res = self.cw.get_dashboard(DashboardName=dashboardname)
            body = json.loads(res['DashboardBody'])
        else:
            with open(CONFIG_DIR + "cloudwatch_dashboard.json", "r") as f:
                body = json.load(f)
        return body

    def init_cloudwatch(self):
        dashboard_name = METRIC_NAMESPACE
        body = self.get_cloudwatch_dashboard(dashboard_name)
        appended = self.add_param_to_widgets(body, METRIC_NAMESPACE)
        body_j = json.dumps(appended)
        self.cw.put_dashboard(DashboardName=dashboard_name,
                              DashboardBody=body_j)

    def setup_update_shadow_lambda(self):
        update_shadow_func = self.name + '_update_shadow'
        api = update_shadow_func + '_api'

        send_mail_topic = self.name + '_send_mail'
        invoke_func = send_mail_topic
        invoke_topic = 'invoke_' + invoke_func
        alarm = self.name + '_alarm'

        # shadowを変更するLambda作成
        self.create_update_shadow_lambda(update_shadow_func, alarm)
        # 上記LambdaをAPI化
        url = self.create_apigateway(api, update_shadow_func)

        # メールを送るSNS作成
        send_mail_topic_arn = self.create_send_mail_sns(send_mail_topic)
        # 上記SNSを呼び出すLambda作成
        func_arn = self.create_send_mail_lambda(invoke_func, send_mail_topic_arn, url)
        # 上記Lambdaを呼び出すSNS作成
        invoke_topic_arn = self.create_sns_invoke_lambda(invoke_topic, invoke_func, func_arn)
        # 上記SNSを呼び出すCloudWatch alarm作成
        self.create_cloudwatch_alarm(alarm, invoke_topic_arn)

    def exists_lambda_function(self, name):
        list_of_funcs = self.lamb.list_functions()
        for func in list_of_funcs['Functions']:
            if func['FunctionName'] == name:
                return True
        while 'NextMarker' in list_of_funcs:
            t = list_of_funcs['NextMarker']
            list_of_funcs = self.lamb.list_functions(Marker=t)
            for func in list_of_funcs['Functions']:
                if func['FunctionName'] == name:
                    return True
        return False

    def create_update_shadow_lambda(self, lambda_func_name, alarm_name):
        # 同名のLambda関数が存在しているか確認
        if self.exists_lambda_function(lambda_func_name):
            return

        # Lambda関数に用いるロールが存在しているか確認。存在しなければ新規作成
        if self.exists_role(lambda_func_name):
            res = self.iam.get_role(RoleName=lambda_func_name)
            role_arn = res['Role']['Arn']
        else:
            role_templ = CONFIG_DIR + "lambda_role.templ"
            policy_templ = CONFIG_DIR + "lambda_update_shadow_policy.templ"
            role_arn = self.create_role(
                name=lambda_func_name,
                roleTempl=role_templ,
                policyTempl=policy_templ
            )

        with open(CONFIG_DIR + "lambda_update_shadow.zip", "rb") as f:
            source = f.read()

        # Lambda関数の作成
        self.lamb.create_function(
            FunctionName=lambda_func_name,
            Runtime="python3.8",
            Role=role_arn,
            Handler="lambda_function.lambda_handler",
            Code={"ZipFile": source},
            Environment={
                'Variables': {
                    'THING_NAME': self.name,
                    'ALARM_NAME': alarm_name
                }
            }
        )

    def exists_rest_api(self, name):
        list_of_apis = self.gw.get_rest_apis()
        for api in list_of_apis['items']:
            if api['name'] == name:
                return True
        while 'position' in list_of_apis:
            t = list_of_apis['position']
            list_of_apis = self.gw.get_rest_apis(position=t)
            for api in list_of_apis['items']:
                if api['name'] == name:
                    return True
        return False

    def get_apiid_resource(self, api_name):
        # 作成したAPIのIDを取得
        apis = {}
        for item in (self.gw.get_rest_apis())['items']:
            apis[item['name']] = item
        apiId = apis[api_name]['id']

        # リソースのIDとPATH取得
        resources = self.gw.get_resources(restApiId=apiId)

        resourceId = resources['items'][0]['id']

        return apiId, resources

    def create_apigateway(self, api_name, lambda_name):
        stage_name = 'prod'
        lambda_arn = self.lamb.get_function(
            FunctionName=lambda_name)['Configuration']['FunctionArn']

        # 同名のAPIが存在しているか確認
        if self.exists_rest_api(api_name):
            # apiId, resoucePathの取得
            apiId, resources = self.get_apiid_resource(api_name)
            resourcePath = resources['items'][0]['path']
            url = 'https://' + apiId + '.execute-api.ap-northeast-1.amazonaws.com/' \
                + stage_name + resourcePath
            return url

        # API作成
        self.gw.create_rest_api(name=api_name)

        # 作成したAPIのIDとリソースを取得
        apiId, resources = self.get_apiid_resource(api_name)

        # リソースのIDとPATH取得
        resources = self.gw.get_resources(restApiId=apiId)
        resourcePath = resources['items'][0]['path']
        resourceId = resources['items'][0]['id']

        # GETメソッド作成
        method = self.gw.put_method(
            restApiId=apiId,
            resourceId=resourceId,
            httpMethod='GET',
            authorizationType='None',
            apiKeyRequired=False
        )

        # Lambda関数がAPI Gatewayから起動することを許可
        self.lamb.add_permission(
            FunctionName=lambda_name,
            StatementId=str(uuid.uuid4()),
            Action='lambda:InvokeFunction',
            Principal='apigateway.amazonaws.com',
            SourceArn='arn:aws:execute-api:ap-northeast-1:' \
            + self.accountId + ':' + apiId + '/*/' \
            + 'GET' + resourcePath
        )

        # 統合リクエストを定義
        with open(CONFIG_DIR + "integrate_request.templ", "r") as f:
            request_template = f.read()
        self.gw.put_integration(
            restApiId=apiId,
            resourceId=resourceId,
            httpMethod='GET',
            type='AWS',
            integrationHttpMethod='POST',
            requestTemplates={
                'application/json': request_template
            },
            uri='arn:aws:apigateway:ap-northeast-1:lambda:path/2015-03-31/functions/' \
            + lambda_arn + '/invocations'
        )

        # メソッドレスポンスを定義
        self.gw.put_method_response(
            restApiId=apiId,
            resourceId=resourceId,
            httpMethod='GET',
            statusCode='200'
        )

        # 統合レスポンスを定義
        self.gw.put_integration_response(
            restApiId=apiId,
            resourceId=resourceId,
            httpMethod='GET',
            statusCode='200',
            responseTemplates={
                'application/json': ''
            }
        )

        # ステージにデプロイ
        self.gw.create_deployment(
            restApiId=apiId,
            stageName=stage_name
        )

        url = 'https://' + apiId + '.execute-api.ap-northeast-1.amazonaws.com/' \
            + stage_name + resourcePath
        return url

    def exists_sns_topic(self, name):
        list_of_topics = self.sns.list_topics()
        for topic in list_of_topics['Topics']:
            if topic['TopicArn'].split(':')[-1] == name:
                return True
        while 'NextToken' in list_of_topics:
            t = list_of_topics['NextToken']
            list_of_topics = self.sns.list_topics(NextToken=t)
            for topic in list_of_topics['Topics']:
                if topic['TopicArn'].split(':')[-1] == name:
                    return True
        return False

    def create_send_mail_sns(self, topic_name):
        #topicの存在確認
        if self.exists_sns_topic(topic_name):
            topic = self.sns.create_topic(Name=topic_name)
            topic_arn = topic['TopicArn']
            return topic_arn

        topic = self.sns.create_topic(Name=topic_name)
        topic_arn = topic['TopicArn']

        # 設定ファイルからメールアドレス取得
        addr_list = []
        with open(CONFIG_DIR + "mail_address_list", "r") as f:
            for i in f:
                if i == '\n':
                    pass
                else:
                    addr_list.append(i.strip())
        # メールアドレスをSubscribeに登録
        for addr in addr_list:
            self.sns.subscribe(
                TopicArn=topic_arn,
                Protocol='email',
                Endpoint=addr
            )
        return topic_arn

    def create_send_mail_lambda(self, lambda_func_name, topic_arn, url):
        # 同名のLambda関数が存在しているか確認
        if self.exists_lambda_function(lambda_func_name):
            res = self.lamb.get_function(FunctionName=lambda_func_name)
            func_arn = res['Configuration']['FunctionArn']
            return func_arn

        # Lambda関数に用いるロールが存在しているか確認。存在しなければ新規作成
        if self.exists_role(lambda_func_name):
            res = self.iam.get_role(RoleName=lambda_func_name)
            role_arn = res['Role']['Arn']
        else:
            role_templ = CONFIG_DIR + "lambda_role.templ"
            policy_templ = CONFIG_DIR + "lambda_send_mail_policy.templ"
            role_arn = self.create_role(
                name=lambda_func_name,
                roleTempl=role_templ,
                policyTempl=policy_templ
            )

        with open(CONFIG_DIR + "lambda_send_mail.zip", "rb") as f:
            source = f.read()

        # Lambda関数の作成
        res = self.lamb.create_function(
            FunctionName=lambda_func_name,
            Runtime="python3.8",
            Role=role_arn,
            Handler="lambda_function.lambda_handler",
            Code={"ZipFile": source},
            Environment={
                'Variables': {
                    'SNS_TOPIC_ARN': topic_arn,
                    'STOP_ALARM_URL': url + "?alert_status=OFF",
                    'START_ALARM_URL': url + "?alert_status=ON"
                }
            }
        )

        return res['FunctionArn']

    def create_sns_invoke_lambda(self, topic_name, func_name, func_arn):
        #topicの存在確認
        if self.exists_sns_topic(topic_name):
            topic = self.sns.create_topic(Name=topic_name)
            topic_arn = topic['TopicArn']
            return topic_arn

        topic = self.sns.create_topic(Name=topic_name)
        topic_arn = topic['TopicArn']

        self.sns.subscribe(
            TopicArn=topic_arn,
            Protocol='lambda',
            Endpoint=func_arn
        )

        # Lambda関数がSNSから起動することを許可
        self.lamb.add_permission(
            FunctionName=func_name,
            StatementId=str(uuid.uuid4()),
            Action='lambda:InvokeFunction',
            Principal='sns.amazonaws.com',
            SourceArn=topic_arn
        )

        return topic_arn

    def create_cloudwatch_alarm(self, alarm_name, topic_arn):
        self.cw.put_metric_alarm(
            AlarmName = alarm_name,
            AlarmActions = [
                topic_arn
            ],
            MetricName=self.name,
            Namespace=METRIC_NAMESPACE,
            Statistic='Maximum',
            Period=60,
            EvaluationPeriods=1,
            Threshold=30,
            ComparisonOperator="GreaterThanThreshold"
        )
